"""
Helper module for date formats.
"""

import datetime


def today() -> str:
    """
    Returns todays date in the correct iso format.
    """
    return datetime.date.isoformat(datetime.date.today())


def is_iso8601(date: str) -> bool:
    """
    Validates that the given date is in the ISO 8601 format (https://en.wikipedia.org/wiki/ISO_8601)
    """
    try:
        datetime.datetime.fromisoformat(date)
        return True
    except ValueError:
        return False



"""
Main module for reading maml files.
"""

import warnings
import os

import yaml
from yaml.parser import ParserError

warnings.formatwarning = lambda msg, *args, **kwargs: f"{msg}\n"


def read_maml(file_name: str) -> dict:
    """
    Reads in a maml file and warns for strange extensions.
    """
    _, extension = os.path.splitext(file_name)
    match extension:
        case ".yml" | ".yaml":
            warnings.warn("WARNING: File was read in but this is a .yml not a .maml\n")
        case ".maml":
            pass
        case _:
            warnings.warn(
                "WARNING: Attempting to read in file but extension is not valid."
            )
    with open(file_name, encoding="utf8") as file:
        try:
            maml_dict = yaml.safe_load(file)
        except ParserError as exc:
            raise ValueError("File is not even valid YAML. Failed to read.") from exc
    if not maml_dict:
        warnings.warn("FILE IS EMPTY!")
        maml_dict = {}
    return maml_dict



# src/pymaml/__init__.py

from .maml import MAML
from .parse import valid_for
from .date_funcs import is_iso8601, today
from .read import read_maml
from .model_v1p0 import V1P0
from .model_v1p1 import V1P1

__all__ = ["MAML", "read_maml", "is_iso8601", "V1P0", "V1P1", "valid_for", "today"]



"""
Main maml object.
"""

import os
import warnings

import yaml
from yaml import SafeDumper

import pandas as pd
from yaml_to_markdown.md_converter import MDConverter

from .read import read_maml
from .parse import MODELS, _assert_version, check_order


def _remove_nones(obj):
    if isinstance(obj, dict):
        return {k: _remove_nones(v) for k, v in obj.items() if v is not None}
    if isinstance(obj, list):
        return [_remove_nones(v) for v in obj if v is not None]
    return obj


class MAML:
    """
    Main MAML class
    """

    def __init__(self, data: dict, version):
        """
        Initializaing and checking that the version is supported
        """
        _assert_version(version)
        self.version = version
        self.meta = MODELS[version](**data)

    @classmethod
    def from_file(cls, file_name: str, version: str) -> "MAML":
        """
        Creates a new maml object from file. Checks that order of keys is in the correct order.
        """
        _assert_version(version)
        data = read_maml(file_name)
        if not check_order(data, version):
            warnings.warn(
                "Ordering is not correct. See (https://github.com/asgr/MAML-Format)"
            )
        return cls(data, version)

    def to_dict(self, include_none: bool = True) -> dict:
        """
        Returns dictionary represenation of the model base class.
        """
        raw = self.meta.model_dump(mode="json")
        if not include_none:
            return _remove_nones(raw)
        return raw

    def to_file(self, file_name: str) -> dict:
        """
        Creates a dictionary representation of the base model.
        """
        SafeDumper.add_representer(
            type(None),
            lambda dumper, _: dumper.represent_scalar("tag:yaml.org,2002:null", ""),
        )
        root, ext = os.path.splitext(file_name)
        if ext != ".maml":
            raise ValueError(f"Extension '{ext}' is not a valid maml extension.")
        with open(f"{root}.maml", "w", encoding="utf8") as file:
            yaml.safe_dump(
                self.to_dict(), file, sort_keys=False, default_flow_style=False
            )

    def to_markdown(self, outfile: str) -> None:
        """
        Dumps the maml as a markdown file.
        """
        data = self.to_dict()
        converter = MDConverter()
        with open(outfile, "w", encoding="utf8") as f:
            converter.convert(data, f)

    def __str__(self) -> str:
        data = self.meta.model_dump(mode="json")
        return f"MAML(version = {self.version}, metadata = {data})"


class MAMLBuilder:
    """
    Builder pattern for constructing the MAML format based on whatever version is decided.
    """

    def __init__(self, version: str, defaults: bool = False):
        """
        Initializing and checking version is valid.
        """
        _assert_version(version)
        self.version: str = version
        self._model_cls = MODELS[version]
        if defaults:
            self._data = self._model_cls.with_defaults().model_dump(mode="json")
        elif not defaults:
            self._data: dict = {}
        else:
            raise ValueError("defaults must be True or False")

    def set(self, field: str, value):
        """
        For setting scalar values
        """
        self._data[field] = value
        return self

    def add(self, field: str, value):
        """
        For adding vector values
        """
        try:
            self._data.setdefault(field, []).append(value)
        except AttributeError:
            self._data[field] = [value]
        return self

    def build(self):
        """
        Attempts to build the class for the current version
        """
        return MAML(self._data, self.version)

    def __str__(self) -> str:
        """
        String represenation showing the current state of the dictionary
        """
        return f"Builder(version = {self.version}, current_build = {self._data})"

    def possible_metadata(self) -> list[str]:
        """
        Lists all the values that can be added in this schema.
        """
        all_values = self._model_cls.with_defaults().model_dump(mode="json").keys()
        return list(all_values)

    def fields_from_pandas(self, pandas_dataframe: pd.DataFrame):
        """
        Fills in the fields from a pandas dataframe using the column names and types.
        """
        field_names = list(pandas_dataframe.columns)
        data_types = [str(dtype) for dtype in list(pandas_dataframe.dtypes)]
        for field_name, dtype in zip(field_names, data_types):
            self.add("fields", {"name": field_name, "data_type": dtype})
        return self



"""
Helper module to parse and check valid maml data structures.
"""

from pydantic_core import ValidationError

from .model_v1p0 import V1P0
from .model_v1p1 import V1P1
from .read import read_maml


MODELS = {
    "v1.0": V1P0,
    "v1.1": V1P1,
}


def _is_subsequence(sub: list, full: list) -> bool:
    """Check if `sub` is a subsequence of `full` (relative order preserved)."""
    it = iter(full)
    return all(any(s == f for f in it) for s in sub)


def check_order(data: dict, version: str) -> bool:
    """
    Recursively check that the order of keys in `data` is a subsequence
    of the schema order defined by the given model version.
    """
    _assert_version(version)
    schema = MODELS[version].with_defaults().model_dump(mode="json")

    def _check_recursive(d: dict, s: dict) -> bool:
        if not isinstance(d, dict) or not isinstance(s, dict):
            return True  # nothing to check here

        d_keys = list(d.keys())
        s_keys = list(s.keys())
        if not _is_subsequence(d_keys, s_keys):
            return False

        for k, dv in d.items():
            if k not in s:
                continue  # ignore unknown keys. Will crash when read in anyways.
            sv = s[k]
            if isinstance(dv, dict) and isinstance(sv, dict):
                if not _check_recursive(dv, sv):
                    return False
            elif isinstance(dv, list) and isinstance(sv, list):
                for d_item, s_item in zip(dv, sv):
                    if isinstance(d_item, dict) and isinstance(s_item, dict):
                        if not _check_recursive(d_item, s_item):
                            return False
        return True

    return _check_recursive(data, schema)


def _assert_version(version: str) -> None:
    """
    Determines if the version is supported and crashes if it isn't.
    """
    if version not in MODELS:
        raise ValueError(
            f"{version} is not a valid version. Supported MAML versions: {list(MODELS.keys())}"
        )


def valid_for(file_name: str) -> list[str]:
    """
    Reads in a file and determines if it is valid for versions of maml.
    """
    dictionary = read_maml(file_name)
    valid = []
    for version, model in MODELS.items():
        try:
            model(**dictionary)
            valid.append(version)
        except ValidationError:
            pass
    if not valid:
        return ["Not valid for any version of MAML"]
    return valid
